from collections import namedtuple


def pRule(peg, name, pf):
    peg[name] = pf

# Generator


def match_empty(px): return True


def pEmpty():
    return match_empty


def pFail():
    return lambda px: False


def match_any(px):
    if px.pos < px.epos:
        px.pos += 1
        return True
    return False


def pAny():
    return match_any


CharCache = {
    '': match_empty
}


def pChar(text):
    if text in CharCache:
        return CharCache[text]
    clen = len(text)

    def match_char(px):
        if px.inputs.startswith(text, px.pos):
            px.pos += clen
            return True
        return False
    CharCache[text] = match_char
    return match_char

# Range


def unique_range(chars, ranges, memo=None):
    cs = 0
    for c in chars:
        cs |= 1 << ord(c)
    r = ranges
    while len(r) > 1:
        for c in range(ord(r[0]), ord(r[1])+1):
            cs |= 1 << c
        r = r[2:]
    if memo is not None:
        if cs in memo:
            return memo[cs]
        memo[cs] = cs
    return cs


def minimum_range(chars, ranges):
    cs = 0xffff
    for c in chars:
        cs = min(cs, ord(c))
    r = ranges
    while len(r) > 1:
        cs = min(cs, ord(r[0]))
        cs = min(cs, ord(r[1]))
        r = r[2:]
    return cs


BitmapCache = {}


def bitmap(chars, ranges):
    key = (chars, ranges)
    if key in BitmapCache:
        return BitmapCache[key]
    offset = minimum_range(chars, ranges)
    bitset = unique_range(chars, ranges) >> offset
    BitmapCache[key] = (bitset, offset)
    return BitmapCache[key]


def pRange(chars, ranges):
    bitset, offset = bitmap(chars, ranges)

    def match_bitset(px):
        if px.pos < px.epos:
            shift = ord(px.inputs[px.pos]) - offset
            if shift >= 0 and (bitset & (1 << shift)) != 0:
                px.pos += 1
                return True
        return False
    return match_bitset


def pAnd(pf):
    def match_and(px):
        pos = px.pos
        if pf(px):
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            return True
        return False
    return match_and


def pNot(pf):
    def match_not(px):
        pos = px.pos
        ast = px.ast
        if not pf(px):
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ast = ast
            return True
        return False
    return match_not


def pMany(pf):
    def match_many(px):
        pos = px.pos
        ast = px.ast
        while pf(px) and pos < px.pos:
            pos = px.pos
            ast = px.ast
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ast = ast
        return True
    return match_many


def pMany1(pf):
    def match_many1(px):
        if pf(px):
            pos = px.pos
            ast = px.ast
            while pf(px) and pos < px.pos:
                pos = px.pos
                ast = px.ast
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ast = ast
            return True
        return False
    return match_many1


def pOption(pf):
    def match_option(px):
        pos = px.pos
        ast = px.ast
        if not pf(px):
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ast = ast
        return True
    return match_option

# Seq


def pSeq2(pf, pf2):
    def match_seq2(px):
        return pf(px) and pf2(px)
    return match_seq2


def pSeq3(pf, pf2, pf3):
    def match_seq3(px):
        return pf(px) and pf2(px) and pf3(px)
    return match_seq3


def pSeq4(pf, pf2, pf3, pf4):
    def match_seq4(px):
        return pf(px) and pf2(px) and pf3(px) and pf4(px)
    return match_seq4


def pSeq(*pfs):
    def match_seq(px):
        for pf in pfs:
            if not pf(px):
                return False
        return True
    return match_seq

# Ore


def pOre2(pf, pf2):
    def match_ore2(px):
        pos = px.pos
        ast = px.ast
        if pf(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ast = ast
        return pf2(px)
    return match_ore2


def pOre3(pf, pf2, pf3):
    def match_ore3(px):
        pos = px.pos
        ast = px.ast
        if pf(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ast = ast
        if pf2(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ast = ast
        return pf3(px)
    return match_ore3


def pOre4(pf, pf2, pf3, pf4):
    def match_ore4(px):
        pos = px.pos
        ast = px.ast
        if pf(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ast = ast
        if pf2(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ast = ast
        if pf3(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ast = ast
        return pf4(px)
    return match_ore4


def pOre(*pfs):
    def match_ore(px):
        pos = px.pos
        ast = px.ast
        for pf in pfs:
            if pf(px):
                return True
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ast = ast
        return False
    return match_ore


def make_trie(dic):
    if '' in dic or len(dic) < 10:
        return dic
    d = {}
    for s in dic:
        s0, s = s[0], s[1:]
        if s0 in d:
            ss = d[s0]
            if not s in ss:
                ss.append(s)
        else:
            d[s0] = [s]
    for key in d:
        d[key] = make_trie(d[key])
    return d


def match_trie(px, d):
    if px.pos >= px.epos:
        return False
    if isinstance(d, dict):
        c = px.inputs[px.pos]
        if c in d:
            px.pos += 1
            return match_trie(px, d[c])
        return False
    pos = px.pos
    inputs = px.inputs
    for s in d:
        if inputs.startswith(s, pos):
            px.pos += len(s)
            return True
    return False


def pDict(words):
    if isinstance(words, str):
        words = words.split(' ')
    dic = make_trie(words)
    return lambda px: match_trie(px, dic)


def pRef(generated, uname):
    if uname not in generated:
        fs = None

        def match_deref(px):
            nonlocal fs
            if fs is None:
                fs = generated[uname]
            return fs(px)
        generated[uname] = match_deref
    return generated[uname]


def pRef(generated, uname):
    if uname not in generated:
        generated[uname] = lambda px: generated[uname](px)
    return generated[uname]


class PMemo(object):
    __slots__ = ['key', 'pos', 'treeState', 'ast', 'prev', 'result']

    def __init__(self):
        self.key = -1
        self.pos = 0
        self.treeState = False
        self.ast = None
        self.prev = None
        self.result = False


def pMemo(fs, mp, mpsize):
    disabled = False
    hit = 0
    miss = 0

    def match_memo(px):
        nonlocal disabled, hit, miss
        if disabled:
            return fs(px)
        key = (mpsize * px.pos) + mp
        m = px.memo[key % 1789]
        if m.key == key:
            if m.treeState:
                if m.prev == px.ast:
                    px.pos = m.pos
                    px.ast = m.ast
                    hit += 1
                    return m.result
            else:
                px.pos = m.pos
                hit += 1
                return m.result
        prev = px.ast
        m.result = fs(px)
        m.pos = px.pos
        m.key = key
        if m.result and prev != px.ast:
            m.treeState = True
            m.prev = prev
            m.ast = px.ast
        else:
            m.treeState = False
        miss += 1
        if miss % 100 == 0:
            if hit / miss < 5:
                disabled = True
        return m.result
    return match_memo


def pMemoDebug(name, fs, mp, mps):
    disabled = False
    hit = 0
    miss = 0
    mpsize = len(mps)

    def match_memo(px):
        nonlocal disabled, hit, miss
        if disabled:
            return fs(px)
        key = (mpsize * px.pos) + mp
        m = px.memo[key % 1789]
        if m.key == key:
            if m.treeState:
                if m.prev == px.ast:
                    px.pos = m.pos
                    px.ast = m.ast
                    hit += 1
                    return m.result
            else:
                px.pos = m.pos
                hit += 1
                return m.result
        prev = px.ast
        m.result = fs(px)
        m.pos = px.pos
        m.key = key
        if m.result and prev != px.ast:
            m.treeState = True
            m.prev = prev
            m.ast = px.ast
        else:
            m.treeState = False
        miss += 1
        if miss % 100 == 0:
            if hit / miss < 5:
                mps.remove(name)
                print('enabled', mps)
                disabled = True
        return m.result
    return match_memo


# Tree Construction


class PTree(object):
    __slots__ = ['prev', 'tag', 'spos', 'epos', 'child']

    def __init__(self, prev, tag, spos, epos, child):
        self.prev = prev
        self.tag = tag
        self.spos = spos
        self.epos = epos
        self.child = child

    def isEdge(self):
        return self.epos < 0

    def dump(self, inputs):
        sb = []
        if self.prev is not None:
            sb.append(self.prev.dump(inputs))
            sb.append(',')
        sb.append(f'{{#{self.tag} ')
        if self.child is None:
            sb.append(repr(inputs[self.spos:self.epos]))
        else:
            sb.append(self.child.dump(inputs))
        sb.append('}')
        return ''.join(sb)


def splitPTree(pt):
    if pt is None:
        return None, None
    if pt.prev is None:
        return None, pt
    return pt.prev, PTree(None, pt.tag, pt.spos, pt.epos, pt.child)


def makePTree(pt: PTree, inputs: str):
    ns = []
    while pt != None:
        if pt.child == None:
            child = repr(inputs[pt.spos:pt.epos])
        else:
            child = makePTree(pt.child)
        child = (pt.tag, child)
        ns.append(child)
        pt = pt.prev
    if len(ns) == 1:
        return ns[0]
    return list(reversed(ns))


def pNode(pf, tag, shift):
    def make_tree(px):
        pos = px.pos
        prev = px.ast
        px.ast = None
        if pf(px):
            px.ast = PTree(prev, tag, pos+shift, px.pos, px.ast)
            return True
        return False
    return make_tree


def pEdge(edge, pf):
    def match_edge(px):
        pos = px.pos
        prev = px.ast
        px.ast = None
        if pf(px):
            px.ast = PTree(prev, edge, pos, -px.pos, px.ast)
            return True
        return False
    return match_edge


def pFold(edge, pf, tag, shift):
    if edge == '':
        def match_fold(px):
            pos = px.pos
            prev, px.ast = splitPTree(px.ast)
            if pf(px):
                px.ast = PTree(prev, tag, pos+shift, px.pos, px.ast)
                return True
            return False
        return match_fold
    else:
        def match_fold2(px):
            pos = px.pos
            # pprev = px.ast
            prev, pt = splitPTree(px.ast)
            px.ast = PTree(None, edge, 0, -pos, pt)
            if pf(px):
                px.ast = PTree(prev, tag, pos+shift, px.pos, px.ast)
                return True
            return False
        return match_fold2


def pAbs(pf):
    def match_abs(px):
        ast = px.ast
        if pf(px):
            px.ast = ast
            return True
        return False
    return match_abs


def pSkip():  # @skip()
    def skip(px):
        px.pos = min(px.headpos, px.epos)
        return True
    return skip

# State


State = namedtuple('State', 'sid val prev')


def getstate(state, sid):
    while state is not None:
        if state.sid == sid:
            return state
        state = state.prev
    return None


def pSymbol(pf, sid):  # @symbol(A)
    def match_symbol(px):
        pos = px.pos
        if pf(px):
            px.state = State(sid, px.inputs[pos:px.pos], px.state)
            return True
        return False
    return match_symbol


def pScope(pf):
    def scope(px):
        state = px.state
        res = pf(px)
        px.state = state
        return res
    return scope


def pExists(sid):  # @Match(A)
    return lambda px: px.getstate(px.state, sid) != None


def pMatch(sid):  # @Match(A)
    def match(px):
        state = px.getstate(px.state, sid)
        if state is not None and px.inputs.startswith(state.val, px.pos):
            px.pos += len(state.val)
            return True
        return False
    return match

#params = pe.params
#name = str(params[0])
#pf = self.emit(pe.e, step)


def pDef(name, pf):
    def define_dic(px):
        pos = px.pos
        if pf(px):
            s = px.inputs[pos:px.pos]
            if len(s) == 0:
                return True
            if name not in px.dic:
                px.dic[name] = []
            ss = px.dic[name]
            ss.append(s)
            px.dic[name] = sorted(ss, key=lambda x: len(x))[::-1]
            # print(px.dic[name])
            return True
        return False
    return define_dic


#params = pe.params
#name = str(params[0])

def pIn(name):  # @in(NAME)
    def match_dic(px):
        #print('@matching', name, px.inputs, px.pos)
        if name in px.dic:
            ss = px.dic[name]
            for s in ss:
                if px.inputs.startswith(s, px.pos):
                    #print('@matched', s)
                    px.pos += len(s)
                    return True
            #print('@', px.inputs[px.pos:], ss)
        return False
    return match_dic


# Optimized


def pAndChar(text):
    def match_andchar(px):
        return px.inputs.startswith(text, px.pos)
    return match_andchar


def pNotChar(text):
    def match_notchar(px):
        return not px.inputs.startswith(text, px.pos)
    return match_notchar


def pManyChar(text):
    clen = len(text)

    def match_manychar(px):
        while px.inputs.startswith(text, px.pos):
            px.pos += clen
        return True
    return match_manychar


def pMany1Char(text):
    clen = len(text)

    def match_many1char(px):
        if px.inputs.startswith(text, px.pos):
            px.pos += clen
            while px.inputs.startswith(text, px.pos):
                px.pos += clen
            return True
        return False
    return match_many1char


def pOptionChar(text):
    clen = len(text)

    def match_optionchar(px):
        if px.inputs.startswith(text, px.pos):
            px.pos += clen
        return True
    return match_optionchar


def pAndRange(chars, ranges):
    bitset, offset = bitmap(chars, ranges)  # >> offset

    def match_andbitset(px):
        if px.pos < px.epos:
            shift = ord(px.inputs[px.pos])  # - offset
            if shift >= 0 and (bitset & (1 << shift)) != 0:
                return True
        return False
    return match_andbitset


def pAndRange(chars, ranges):
    bitset, offset = bitmap(chars, ranges)  # >> offset

    def match_andbitset(px):
        if px.pos < px.epos:
            shift = ord(px.inputs[px.pos]) - offset
            if shift >= 0 and (bitset & (1 << shift)) != 0:
                return True
        return False
    return match_andbitset


def pNotRange(chars, ranges):
    bitset, offset = bitmap(chars, ranges)  # >> offset

    def match_notbitset(px):
        if px.pos < px.epos:
            shift = ord(px.inputs[px.pos]) - offset
            if shift >= 0 and (bitset & (1 << shift)) != 0:
                return False
        return True
    return match_notbitset


def pManyRange(chars, ranges):
    bitset, offset = bitmap(chars, ranges)  # >> offset

    def match_manybitset(px):
        while px.pos < px.epos:
            shift = ord(px.inputs[px.pos]) - offset
            if shift >= 0 and (bitset & (1 << shift)) != 0:
                px.pos += 1
            else:
                break
        return True
    return match_manybitset


def pMany1Range(chars, ranges):
    return pSeq2(pRange(chars, ranges), pManyRange(chars, ranges))


def pOptionRange(chars, ranges):
    bitset, offset = bitmap(chars, ranges)  # >> offset

    def match_optionbitset(px):
        if px.pos < px.epos:
            shift = ord(px.inputs[px.pos]) - offset
            if shift >= 0 and (bitset & (1 << shift)) != 0:
                px.pos += 1
        return True
    return match_optionbitset

# generate


# PContext


class PContext:
    __slots__ = ['inputs', 'pos', 'epos',
                 'headpos', 'ast', 'state', 'memo', 'dic']

    def __init__(self, inputs, spos, epos):
        self.inputs = inputs
        self.pos = spos
        self.epos = epos
        self.headpos = spos
        self.ast = None
        self.state = None
        self.memo = [PMemo() for x in range(1789)]
        self.dic = {}

# ParseTree


def rowcol(urn, inputs, spos):
    inputs = inputs[:spos + (1 if len(inputs) > spos else 0)]
    rows = inputs.split(b'\n' if isinstance(inputs, bytes) else '\n')
    return urn, spos, len(rows), len(rows[-1])-1


def nop(s): return s


UNKNOWN_SOURCE = '(unknown source)'


class ParseTree(list):
    def __init__(self, tag, inputs, spos=0, epos=None, urn=UNKNOWN_SOURCE):
        self.tag_ = tag
        self.inputs_ = inputs
        self.spos_ = spos
        self.epos_ = epos if epos is not None else len(inputs)
        self.urn_ = urn

    def gettag(self):
        return self.tag_

    def start(self):
        return rowcol(self.urn_, self.inputs_, self.spos_)

    def end(self):
        return rowcol(self.urn_, self.inputs_, self.epos_)

    def decode(self):
        inputs, spos, epos = self.inputs_, self.spos_, self.epos_
        LF = b'\n' if isinstance(inputs, bytes) else '\n'
        rows = inputs[:spos + (1 if len(inputs) > spos else 0)]
        rows = rows.split(LF)
        linenum, column = len(rows), len(rows[-1])-1
        begin = inputs.rfind(LF, 0, spos) + 1
        # print('@', spos, begin, inputs)
        end = inputs.find(LF, spos)
        # print('@', spos, begin, inputs)
        if end == -1:
            end = len(inputs)
        # print('@[', begin, spos, end, ']', epos)
        line = inputs[begin:end]  # .replace('\t', '   ')
        mark = []
        endcolumn = column + (epos - spos)
        for i, c in enumerate(line):
            if column <= i and i <= endcolumn:
                mark.append('^' if ord(c) < 256 else '^^')
            else:
                mark.append(' ' if ord(c) < 256 else '  ')
        mark = ''.join(mark)
        return (self.urn_, spos, linenum, column, line, mark)

    def showing(self, msg='Syntax Error'):
        urn, pos, linenum, cols, line, mark = self.decode()
        return '{} ({}:{}:{}+{})\n{}\n{}'.format(msg, urn, linenum, cols, pos, line, mark)

    def __eq__(self, tag):
        return self.tag_ == tag

    def isSyntaxError(self):
        return self.tag_ == 'err'

    def __str__(self):
        s = self.inputs_[self.spos_:self.epos_]
        return s.decode('utf-8') if isinstance(s, bytes) else s

    def __repr__(self):
        if self.isSyntaxError():
            return self.showing('Syntax Error')
        sb = []
        self.strOut(sb)
        return "".join(sb)

    def dump(self, indent='\n  ', tab='  ', tag=nop, edge=nop, token=nop):
        if self.isSyntaxError():
            print(self.showing('Syntax Error'))
        else:
            sb = []
            self.strOut(sb)
            print("".join(sb))

    def strOut(self, sb, indent='\n  ', tab='  ', tag=nop, edge=nop, token=nop):
        sb.append("[" + tag(f'#{self.tag_}'))
        hasContent = False
        next_indent = indent + tab
        for child in self:
            hasContent = True
            sb.append(indent)
            if hasattr(child, 'strOut'):
                child.strOut(sb, next_indent, tab, tag, edge, token)
            else:
                sb.append(repr(child))
        for key in self.__dict__:
            v = self.__dict__[key]
            if isinstance(v, ParseTree):
                hasContent = True
                sb.append(indent)
                sb.append(edge(key) + ': ')
                v.strOut(sb, next_indent, tab, tag, edge, token)
        if not hasContent:
            sb.append(' ' + token(repr(str(self))))
        sb.append("]")


def PTree2ParseTree(pt: PTree, urn, inputs):
    if pt.prev != None:
        return PTree2ParseTreeImpl('', urn, inputs, pt.spos, pt.epos, pt)
    else:
        return PTree2ParseTreeImpl(pt.tag, urn, inputs, pt.spos, pt.epos, pt.child)


def PTree2ParseTreeImpl(tag, urn, inputs, spos, epos, subnode):
    t = ParseTree(tag, inputs, spos, epos, urn)
    while subnode != None:
        if subnode.isEdge():
            if subnode.child == None:
                tt = PTree2ParseTreeImpl(
                    '', urn, inputs, subnode.spos, abs(subnode.epos), None)
            else:
                tt = PTree2ParseTree(subnode.child, urn, inputs)
            if subnode.tag == '':
                t.append(tt)
            else:
                setattr(t, subnode.tag, tt)
        else:
            t.append(PTree2ParseTreeImpl(subnode.tag, urn, inputs,
                                         subnode.spos, abs(subnode.epos), subnode.child))
        subnode = subnode.prev
    for i in range(len(t)//2):
        t[i], t[-(1+i)] = t[-(1+i)], t[i]
    return t


def generate(pf):
    # pf = self.generated[start.uname()]
    def parse(inputs, urn='(unknown source)', pos=0, epos=None, conv=PTree2ParseTree):
        if epos is None:
            epos = len(inputs)
        px = PContext(inputs, pos, epos)
        if not pf(px):
            result = PTree(None, "err", px.headpos, px.headpos, None)
        else:
            result = px.ast if px.ast is not None else PTree(None,
                                                             "", pos, px.pos, None)
        return conv(result, urn, inputs)
    return parse


# TPEG



import unittest
from test_gpeg import TestGPEG
from test_tpeg import TestTPEG

def suite():
  suite = unittest.TestSuite()
  suite.addTests(unittest.makeSuite(TestGPEG))
  suite.addTests(unittest.makeSuite(TestTPEG))
  return suite



# -*- coding: utf-8 -*-
import sys
import os
import errno
import inspect
from pathlib import Path
import pegtree.pasm as pasm
from pegtree.tpeg import TPEGGrammar
# sys.setrecursionlimit(5000)


def DEBUG(*x):
    if 'DEBUG' in os.environ:
        print('DEBUG', *x)

# ParsingExpression


class PExpr(object):
    def __iter__(self): pass
    def __len__(self): return 0

    def cname(self):
        return self.__class__.__name__


class PAny(PExpr):
    def __repr__(self): return '.'
    def minLen(self): return 1


class PChar(PExpr):
    __slots__ = ['text']
    ESCTBL = str.maketrans(
        {'\n': '\\n', '\t': '\\t', '\r': '\\r', '\v': '\\v', '\f': '\\f',
         '\\': '\\\\', "'": "\\'"})

    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return "'" + self.text.translate(PChar.ESCTBL) + "'"

    def minLen(self): return len(self.text)


# def newranges(ranges):
#     if not isinstance(ranges, str):
#         sb = []
#         for r in ranges:
#             sb.append(r[0]+r[1])
#         return ''.join(sb)
#     return ranges


class PRange(PExpr):
    __slots__ = ['chars', 'ranges']
    ESCTBL = str.maketrans(
        {'\n': '\\n', '\t': '\\t', '\r': '\\r', '\v': '\\v', '\f': '\\f',
         '\\': '\\\\', ']': '\\]', '-': '\\-'})

    def __init__(self, chars, ranges):
        self.chars = chars
        self.ranges = ranges  # newranges(ranges)

    def __repr__(self):
        sb = []
        sb.append('[')
        sb.append(self.chars.translate(PRange.ESCTBL))
        r = self.ranges
        while len(r) > 1:
            sb.append(r[0].translate(PRange.ESCTBL))
            sb.append('-')
            sb.append(r[1].translate(PRange.ESCTBL))
            r = r[2:]
        sb.append(']')
        return ''.join(sb)

    def minLen(self): return 1


def isSingleChar(e):
    return (isinstance(e, PChar) and len(e.text) == 1) or isinstance(e, PRange)


def mergeRange(e, e2):
    chars = ''
    ranges = ''
    if isinstance(e, PChar):
        chars += e.text
    if isinstance(e2, PChar):
        chars += e2.text
    if isinstance(e, PRange):
        chars += e.chars
        ranges += e.ranges
    if isinstance(e2, PRange):
        chars += e2.chars
        ranges += e2.ranges
    return PRange(chars, ranges)


class PRef(PExpr):
    def __init__(self, peg, name):
        self.peg = peg
        self.name = name

    def __repr__(self):
        return self.name

    def uname(self, peg=None):
        if self.peg == peg:
            return self.name
        return f'{self.peg.ns}{self.name}'

    def deref(self):
        return self.peg[self.name]

    def minLen(self):
        if not hasattr(self, 'minlen'):
            self.minlen = 0
            self.minlen = self.deref().minLen()
        return self.minlen


class PTuple(PExpr):
    __slots__ = ['es']

    def __init__(self, *es):
        self.es = list(es)

    def __iter__(self):
        return iter(self.es)

    def __len__(self):
        return len(self.es)

    def minLen(self):
        if not hasattr(self, 'minlen'):
            self.minlen = min(map(lambda e: e.minLen(), self.es))
        return self.minlen


class PSeq(PTuple):
    @classmethod
    def new(cls, *es):
        if len(es) == 0:
            return EMPTY
        ls = [es[0]]
        for e in es[1:]:
            if e == EMPTY:
                continue
            if isinstance(e, PChar) and isinstance(ls[-1], PChar):
                ls[-1] = PChar(ls[-1].text+e.text)
                continue
            ls.append(e)
        return ls[0] if len(ls) == 1 else PSeq(*ls)

    def __repr__(self):
        return ' '.join(map(ss, self))

    def minLen(self):
        if not hasattr(self, 'minlen'):
            self.minlen = sum(map(lambda e: e.minLen(), self.es))
        return self.minlen


class PAlt(PTuple):
    def __repr__(self):
        return ' | '.join(map(repr, self))


class POre(PTuple):
    @classmethod
    def new(cls, *es):
        choices = []
        for e in es:
            appendChoice(choices, e)
        return choices[0] if len(choices) == 1 else POre(*choices)

    def __repr__(self):
        return ' / '.join(map(repr, self))

    def optimize(self):
        choices = []
        optimizedChoice(choices, self)
        return choices[0] if len(choices) == 1 else POre(*choices)

    def isDict(self):
        for e in self:
            if not isinstance(e, PChar):
                return False
        return True

    def listDict(self):
        dic = [e.text for e in self if isinstance(e, PChar)]
        dic2 = []
        for s in dic:
            if s == '':
                break
            dic2.append(s)
        return dic2


def appendChoice(choices, pe):
    if isinstance(pe, POre):
        for e in pe:
            appendChoice(choices, e)
    elif len(choices) == 0:
        choices.append(pe)
    elif isSingleChar(choices[-1]) and isSingleChar(pe):
        DEBUG('OPTIMIZE', choices[-1], pe, '=>', mergeRange(choices[-1], pe))
        choices[-1] = mergeRange(choices[-1], pe)
    elif choices[-1] != EMPTY:
        choices.append(pe)
    else:
        DEBUG('IGNORED', pe)


def optimizedChoice(choices, pe):
    start = pe
    while isinstance(pe, PRef):
        pe = pe.deref()
    if isinstance(pe, POre):
        if(pe != start):
            DEBUG('INLINE', start, '=>', pe)
        for e in pe:
            optimizedChoice(choices, e)
    elif isinstance(pe, PRange) or isinstance(pe, PChar):
        if(pe != start):
            DEBUG('INLINE', start, '=>', pe)
        appendChoice(choices, pe)
    else:
        appendChoice(choices, start)


class PUnary(PExpr):
    __slot__ = ['e']

    def __init__(self, e):
        self.e = e

    def __iter__(self):
        yield self.e

    def __len__(self):
        return 1

    def minLen(self): return self.e.minLen()


class PAnd(PUnary):
    def __repr__(self):
        return '&'+grouping(self.e, inUnary)

    def minLen(self): return 0


class PNot(PUnary):
    def __repr__(self):
        return '!'+grouping(self.e, inUnary)

    def minLen(self): return 0


class PMany(PUnary):
    def __repr__(self):
        return grouping(self.e, inUnary)+'*'

    def minLen(self): return 0


class PMany1(PUnary):
    def __repr__(self):
        return grouping(self.e, inUnary)+'+'


class POption(PUnary):
    def __repr__(self):
        return grouping(self.e, inUnary)+'?'

    def minLen(self): return 0


class PNode(PUnary):
    __slot__ = ['e', 'tag', 'shift']

    def __init__(self, e, tag='', shift=0):
        self.e = e
        self.tag = tag
        self.shift = shift

    def __repr__(self):
        s = '' if self.shift == 0 else f'/*{self.shift}*/'
        return '{' + s + repr(self.e) + ' #' + self.tag + '}'


class PEdge(PUnary):
    __slot__ = ['e', 'edge']

    def __init__(self, edge, e):
        self.e = e
        self.edge = edge

    def __repr__(self):
        return self.edge + ': ' + grouping(self.e, inUnary)


class PFold(PUnary):
    __slot__ = ['e', 'edge', 'tag', 'shift']

    def __init__(self, edge, e, tag='', shift=0):
        self.e = e
        self.edge = edge
        self.tag = tag
        self.shift = shift

    def __repr__(self):
        s = '' if self.shift == 0 else f'/*{self.shift}*/'
        if self.edge == '':
            return '^ {' + s + repr(self.e) + ' #' + self.tag + '}'
        return self.edge + ': ^ {' + s + repr(self.e) + ' #' + self.tag + '}'


class PAbs(PUnary):
    __slot__ = ['e']

    def __init__(self, e):
        self.e = e

    def __repr__(self):
        return f'@abs({self.e})'

# Action


class PAction(PUnary):
    __slots__ = ['e', 'func', 'params']

    def __init__(self, e, func, params, pos4=None):
        self.e = e
        self.func = func
        self.params = params

    def __repr__(self):
        return f'@{self.func}{self.params}'

    def cname(self):
        return self.func.capitalize()


# CONSTANT
EMPTY = PChar('')
ANY = PAny()
FAIL = PNot(EMPTY)

'''
def pEmpty(): return EMPTY


def pAny(): return ANY


def pChar(c: str): return PChar(c) if len(c) > 0 else EMPTY


def pRange(cs: str, rs: str): return PRange(cs, rs)


def pAnd(e: PExpr): return PAnd(e)


def pNot(e: PExpr): return PNot(e)


def pMany(e: PExpr): return PMany(e)


def pMany1(e: PExpr): return PMany1(e)


def pOption(e: PExpr): return POption(e)


def pSeq(*es): return PSeq(*es)


def pSeq2(e: PExpr, e2: PExpr): return PSeq(e, e2)


def pSeq3(e: PExpr, e2: PExpr, e3: PExpr): return PSeq(e, e2, e3)


def pOre(*es): return POre.new(*es)


def pOre2(e: PExpr, e2: PExpr): return POre.new(e, e2)


def pOre3(e: PExpr, e2: PExpr, e3: PExpr): return POre.new(e, e2, e3)


def pRef(peg, name: str): return PRef(peg, name)


def pNode(e, tag, shift): return PNode(e, tag)


def pEdge(label, e): return PEdge(label, e) if label != '' else e


def pFold(label, e, tag, shift): return PFold(label, e, tag)
'''

# repr


def grouping(e, f):
    return '(' + repr(e) + ')' if f(e) else repr(e)


def inUnary(e):
    return isinstance(e, POre) \
        or isinstance(e, PSeq) or isinstance(e, PAlt) \
        or (isinstance(e, PEdge))or isinstance(e, PFold)


def ss(e):
    return grouping(e, lambda e: isinstance(e, POre) or isinstance(e, PAlt))

# # Grammar


GrammarId = 0


class Grammar(dict):
    def __init__(self, source=None):
        global GrammarId
        self.ns = str(GrammarId)
        self.N = []
        GrammarId += 1
        super().__setitem__('@@example', [])

    def __repr__(self):
        ss = []
        for rule in self.N:
            ss.append(rule)
            ss.append(' = ')
            ss.append(repr(self[rule]))
            ss.append('\n')
        return ''.join(ss)

    # def add(self, key, item):
    #     if not key in self:
    #         self.N.append(key)
    #     self[key] = item

    def __setitem__(self, key, item):
        if not key in self:
            self.N.append(key)
        super().__setitem__(key, item)

    def newRef(self, name):
        key = '@' + name
        if key not in self:
            super().__setitem__(key, PRef(self, name))
        return self[key]

    def start(self):
        if len(self.N) == 0:
            self['EMPTY'] = EMPTY
        return self.N[0]


# def pRule(peg, name, pf):
#     peg[name] = pf


# def TPEG(peg):
#     pRule(peg, 'Start', pSeq3(pRef(peg, '__'), pRef(
#         peg, 'Source'), pRef(peg, 'EOF')))
#     pRule(peg, '__', pMany(pOre2(pRange(' \t\r\n', []), pRef(peg, 'COMMENT'))))
#     pRule(peg, '_', pMany(pOre2(pRange(' \t', []), pRef(peg, 'COMMENT'))))
#     pRule(peg, 'COMMENT', pOre2(pSeq3(pChar('/*'), pMany(pSeq2(pNot(pChar('*/')), pAny())),
#                                       pChar('*/')), pSeq2(pChar('//'), pMany(pSeq2(pNot(pRef(peg, 'EOL')), pAny())))))
#     pRule(peg, 'EOL', pOre(pChar('\n'), pChar('\r\n'), pRef(peg, 'EOF')))
#     pRule(peg, 'EOF', pNot(pAny()))
#     pRule(peg, 'S', pRange(' \t', []))
#     pRule(peg, 'Source', pNode(
#         pMany(pEdge('', pRef(peg, 'Statement'))), 'Source', 0))
#     pRule(peg, 'EOS', pOre2(pSeq2(pRef(peg, '_'), pMany1(pSeq2(pChar(';'), pRef(
#         peg, '_')))), pMany1(pSeq2(pRef(peg, '_'), pRef(peg, 'EOL')))))
#     pRule(peg, 'Statement', pOre(pRef(peg, 'Import'), pRef(
#         peg, 'Example'), pRef(peg, 'Rule')))
#     pRule(peg, 'Import', pSeq2(pNode(pSeq(pChar('from'), pRef(peg, 'S'), pRef(peg, '_'), pEdge('name', pOre2(pRef(peg, 'Identifier'), pRef(peg, 'Char'))), pOption(
#         pSeq(pRef(peg, '_'), pChar('import'), pRef(peg, 'S'), pRef(peg, '_'), pEdge('names', pRef(peg, 'Names'))))), 'Import', 0), pRef(peg, 'EOS')))
#     pRule(peg, 'Example', pSeq2(pNode(pSeq(pChar('example'), pRef(peg, 'S'), pRef(peg, '_'), pEdge(
#         'names', pRef(peg, 'Names')), pEdge('doc', pRef(peg, 'Doc'))), 'Example', 0), pRef(peg, 'EOS')))
#     pRule(peg, 'Names', pNode(pSeq3(pEdge('', pRef(peg, 'Identifier')), pRef(peg, '_'), pMany(pSeq(
#         pChar(','), pRef(peg, '_'), pEdge('', pRef(peg, 'Identifier')), pRef(peg, '_')))), '', 0))
#     pRule(peg, 'Doc', pOre(pRef(peg, 'Doc1'),
#                            pRef(peg, 'Doc2'), pRef(peg, 'Doc0')))
#     pRule(peg, 'Doc0', pNode(pMany(pSeq2(pNot(pRef(peg, 'EOL')), pAny())), 'Doc', 0))
#     pRule(peg, 'Doc1', pSeq(pRef(peg, 'DELIM1'), pMany(pRef(peg, 'S')), pRef(peg, 'EOL'), pNode(pMany(
#         pSeq2(pNot(pSeq2(pRef(peg, 'DELIM1'), pRef(peg, 'EOL'))), pAny())), 'Doc', 0), pRef(peg, 'DELIM1')))
#     pRule(peg, 'DELIM1', pChar("'''"))
#     pRule(peg, 'Doc2', pSeq(pRef(peg, 'DELIM2'), pMany(pRef(peg, 'S')), pRef(peg, 'EOL'), pNode(pMany(
#         pSeq2(pNot(pSeq2(pRef(peg, 'DELIM2'), pRef(peg, 'EOL'))), pAny())), 'Doc', 0), pRef(peg, 'DELIM2')))
#     pRule(peg, 'DELIM2', pChar('```'))
#     pRule(peg, 'Rule', pSeq2(pNode(pSeq(pEdge('name', pOre2(pRef(peg, 'Identifier'), pRef(peg, 'QName'))), pRef(peg, '__'), pOre2(pChar('='), pChar(
#         '<-')), pRef(peg, '__'), pOption(pSeq2(pRange('/|', []), pRef(peg, '__'))), pEdge('e', pRef(peg, 'Expression'))), 'Rule', 0), pRef(peg, 'EOS')))
#     pRule(peg, 'Identifier', pNode(pRef(peg, 'NAME'), 'Name', 0))
#     pRule(peg, 'NAME', pSeq2(pRange('_', ['AZ', 'az']),
#                              pMany(pRange('_.', ['AZ', 'az', '09']))))
#     pRule(peg, 'Expression', pSeq2(pRef(peg, 'Choice'), pOption(pFold('', pMany1(pSeq(pRef(peg, '__'), pChar(
#         '|'), pNot(pChar('|')), pRef(peg, '_'), pEdge('', pRef(peg, 'Choice')))), 'Alt', 0))))
#     pRule(peg, 'Choice', pSeq2(pRef(peg, 'Sequence'), pOption(pFold('', pMany1(pSeq(pRef(peg, '__'), pOre2(
#         pChar('/'), pChar('||')), pRef(peg, '_'), pEdge('', pRef(peg, 'Sequence')))), 'Ore', 0))))
#     pRule(peg, 'Sequence', pSeq2(pRef(peg, 'Predicate'), pOption(pFold('', pMany1(
#         pSeq2(pRef(peg, 'SS'), pEdge('', pRef(peg, 'Predicate')))), 'Seq', 0))))
#     pRule(peg, 'SS', pOre2(pSeq3(pRef(peg, 'S'), pRef(peg, '_'), pNot(pRef(peg, 'EOL'))), pSeq3(
#         pMany1(pSeq2(pRef(peg, '_'), pRef(peg, 'EOL'))), pRef(peg, 'S'), pRef(peg, '_'))))
#     pRule(peg, 'Predicate', pOre(pRef(peg, 'Not'), pRef(
#         peg, 'And'), pRef(peg, 'Suffix')))
#     pRule(peg, 'Not', pSeq2(pChar('!'), pNode(
#         pEdge('e', pRef(peg, 'Predicate')), 'Not', 0)))
#     pRule(peg, 'And', pSeq2(pChar('&'), pNode(
#         pEdge('e', pRef(peg, 'Predicate')), 'And', 0)))
#     pRule(peg, 'Suffix', pSeq2(pRef(peg, 'Term'), pOption(pOre(pFold('e', pChar(
#         '*'), 'Many', 0), pFold('e', pChar('+'), 'Many1', 0), pFold('e', pChar('?'), 'Option', 0)))))
#     pRule(peg, 'Term', pOre(pRef(peg, 'Group'), pRef(peg, 'Char'), pRef(peg, 'Class'), pRef(peg, 'Any'), pRef(
#         peg, 'Node'), pRef(peg, 'Fold'), pRef(peg, 'EdgeFold'), pRef(peg, 'Edge'), pRef(peg, 'Func'), pRef(peg, 'Ref')))
#     pRule(peg, 'Empty', pNode(pEmpty(), 'Empty', 0))
#     pRule(peg, 'Group', pSeq(pChar('('), pRef(peg, '__'), pOre2(
#         pRef(peg, 'Expression'), pRef(peg, 'Empty')), pRef(peg, '__'), pChar(')')))
#     pRule(peg, 'Any', pNode(pChar('.'), 'Any', 0))
#     pRule(peg, 'Char', pSeq3(pChar("'"), pNode(pMany(pOre2(pSeq2(
#         pChar('\\'), pAny()), pSeq2(pNot(pChar("'")), pAny()))), 'Char', 0), pChar("'")))
#     pRule(peg, 'Class', pSeq3(pChar('['), pNode(pMany(pOre2(pSeq2(
#         pChar('\\'), pAny()), pSeq2(pNot(pChar(']')), pAny()))), 'Class', 0), pChar(']')))
#     pRule(peg, 'Node', pNode(pSeq(pChar('{'), pRef(peg, '__'), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(peg, '__'))), pEdge('e', pOre2(pSeq2(pRef(
#         peg, 'Expression'), pRef(peg, '__')), pRef(peg, 'Empty'))), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(peg, '__'))), pRef(peg, '__'), pChar('}')), 'Node', 0))
#     pRule(peg, 'Tag', pSeq2(pChar('#'), pNode(
#         pMany1(pSeq2(pNot(pRange(' \t\r\n}', [])), pAny())), 'Tag', 0)))
#     pRule(peg, 'Fold', pNode(pSeq(pChar('^'), pRef(peg, '_'), pChar('{'), pRef(peg, '__'), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(peg, '__'))), pEdge('e', pOre2(pSeq2(
#         pRef(peg, 'Expression'), pRef(peg, '__')), pRef(peg, 'Empty'))), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(peg, '__'))), pRef(peg, '__'), pChar('}')), 'Fold', 0))
#     pRule(peg, 'Edge', pNode(pSeq(pEdge('edge', pRef(peg, 'Identifier')), pChar(':'), pRef(
#         peg, '_'), pNot(pChar('^')), pEdge('e', pRef(peg, 'Term'))), 'Edge', 0))
#     pRule(peg, 'EdgeFold', pNode(pSeq(pEdge('edge', pRef(peg, 'Identifier')), pChar(':'), pRef(peg, '_'), pChar('^'), pRef(peg, '_'), pChar('{'), pRef(peg, '__'), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(
#         peg, '__'))), pEdge('e', pOre2(pSeq2(pRef(peg, 'Expression'), pRef(peg, '__')), pRef(peg, 'Empty'))), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(peg, '__'))), pRef(peg, '__'), pChar('}')), 'Fold', 0))
#     pRule(peg, 'Func', pNode(pSeq(pChar('@'), pEdge('', pRef(peg, 'Identifier')), pChar('('), pRef(peg, '__'), pOre2(pEdge('', pRef(peg, 'Expression')), pEdge(
#         '', pRef(peg, 'Empty'))), pMany(pSeq(pRef(peg, '_'), pChar(','), pRef(peg, '__'), pEdge('', pRef(peg, 'Expression')))), pRef(peg, '__'), pChar(')')), 'Func', 0))
#     pRule(peg, 'Ref', pOre2(pRef(peg, 'Identifier'), pRef(peg, 'QName')))
#     pRule(peg, 'QName', pNode(pSeq3(pChar('"'), pMany(pOre2(pSeq2(
#         pChar('\\'), pAny()), pSeq2(pNot(pChar('"')), pAny()))), pChar('"')), 'Name', 0))
#     return peg


# TPEGGrammar = TPEG(Grammar())
# print(TPEGGrammar)

######################################################################
# ast.env

class Optimizer(object):

    def inline(self, pe: PExpr):
        start = pe
        while isinstance(pe, PRef):
            pe = pe.deref()
        if isinstance(pe, PChar) or isinstance(pe, PRange):
            if(pe != start):
                DEBUG('INLINE', start, '=>', pe)
            return pe
        return start

    def join(self, *es):
        return PSeq.new(*es)

    def fixedExpr(self, e):
        es = [e]
        fixed = []
        size = 0
        while len(es) > 0:
            # print(es, '->')
            lsize, e, es = self.fixedEach(size, es)
            # print('->', lsize, e, es)
            if e is None:
                break
            fixed.append(e)
            size += lsize
        return size, fixed, es

    def fixedEach(self, size, es):
        e = self.inline(es[0])
        if isinstance(e, PChar) and len(e.text) > 0:
            return len(e.text)+size, e, es[1:]
        elif isinstance(e, PRange) or isinstance(e, PAny):
            return 1+size, e, es[1:]
        elif isinstance(e, PAnd) or isinstance(e, PNot):
            return size, e, es[1:]
        elif isinstance(e, PSeq):
            return self.fixedEach(size, e.es+es[1:])
        elif isinstance(e, PMany1):
            lsize, lfixed, les = self.fixedExpr(e.e)
            # print('PMany1', e, '=>', lsize, lfixed, les)
            if len(les) != 0:
                return size, None, es
            return size+lsize, e.e, [PMany(e.e)] + es[1:]
        elif isinstance(e, PNode) and e.shift == 0:
            lsize, lfixed, les = self.fixedExpr(e.e)
            # print('PNode', e, '=>', lsize, lfixed, les)
            if len(lfixed) == 0:
                return size, None, es
            return size+lsize, PSeq.new(*lfixed), [PNode(PSeq.new(*les), e.tag, -lsize)]+es[1:]
        elif isinstance(e, PFold) and e.shift == 0:
            lsize, lfixed, les = self.fixedExpr(e.e)
            if len(lfixed) == 0:
                return size, None, es
            return size+lsize, PSeq.new(*lfixed), [PFold(e.edge, PSeq.new(*les), e.tag, -lsize)]+es[1:]
        return size, None, es

    def sort(self, refs):
        newrefs = []
        unsolved = []
        for ref in refs:
            names = set([])
            self.makerefs(ref.deref(), names)
            if len(names) == 0:
                newrefs.append(ref)
            else:
                unsolved.append((ref, set(names)))
        return self.solve(newrefs, unsolved)

    def makerefs(self, e, names):
        if isinstance(e, PTuple):
            for e2 in e:
                self.makerefs(e2, names)
        elif hasattr(e, 'e'):
            self.makerefs(e.e, names)
        elif isinstance(e, PRef):
            names.add(e.uname())

    def removeSolvedName(self, unsolved, uname):
        removed = False
        for _, names in unsolved:
            if uname in names:
                removed = True
                names.remove(uname)
        return removed

    def solve(self, refs, unsolved):
        removed = False
        # print(refs)
        for ref in refs:
            removed |= self.removeSolvedName(unsolved, ref.uname())
        max = 0
        while max < 10:
            removed = True
            while removed:
                removed = False
                newrefs = []
                stillUnsolved = []
                for ref, names in unsolved:
                    if len(names) <= max:
                        refs.append(ref)
                        newrefs.append(ref)
                    else:
                        stillUnsolved.append((ref, names))
                unsolved = stillUnsolved
                #print(max, newrefs)
                for ref in newrefs:
                    removed |= self.removeSolvedName(unsolved, ref.uname())
                if removed:
                    max = 0
            max += 1
        for ref, _ in unsolved:
            refs.append(ref)
        return refs


class Generator(Optimizer):
    def __init__(self):
        self.peg = None
        self.generated = {}
        self.generating_nonterminal = ''
        self.sids = {}
        self.memos = []
        self.Ooox = True
        self.Olex = True

    def getsid(self, name):
        if not name in self.sids:
            self.sids[name] = len(self.sids)
        return self.sids[name]

    def makelist(self, pe, v: dict, ps: list):
        if isinstance(pe, PRef):
            u = pe.uname()
            if u not in self.generated and u not in v:
                v[u] = pe
                self.makelist(pe.deref(), v, ps)
                ps.append(pe)
            return ps
        if isinstance(pe, PUnary) or isinstance(pe, PTuple):
            for e in pe:
                self.makelist(e, v, ps)
        return ps

    def generate(self, peg, **option):
        self.peg = peg
        name = option.get('start', peg.start())
        start = peg.newRef(name)
        # if 'memos' in option and not isinstance(option['memos'], list):
        if 'packrat' in peg:
            memos = peg['packrat']
            if isinstance(memos, POre):
                self.memos = memos.listDict()
            else:
                self.memos = peg.N
            # print(self.memos)
        ps = self.makelist(start, {}, [])
        ps = self.sort(ps)
        for ref in ps:
            assert isinstance(ref, PRef)
            self.generating_nonterminal = ref.uname()
            self.emitRule(ref)
            self.generating_nonterminal = ''

        return self.emitParser(start)

    def emitRule(self, ref):
        A = self.emit(ref.deref(), 0)
        if ref.peg == self.peg and ref.name in self.memos:
            idx = self.memos.index(ref.name)
            if idx != -1:
                A = pasm.pMemo(A, idx, len(self.memos))
                # A = pasm.pMemoDebug(ref.name, A, idx, self.memos)
        self.generated[ref.uname()] = A

    def emitParser(self, start):
        return pasm.generate(self.generated[start.uname()])

    def emit(self, pe: PExpr, step: int):
        pe = self.inline(pe)
        cname = pe.cname()
        if hasattr(self, cname):
            f = getattr(self, cname)
            return f(pe, step)
        print('@TODO(Generator)', cname, pe)
        return self.PChar(EMPTY, step)

    def PAny(self, pe, step):
        return pasm.pAny()

    def PChar(self, pe, step):
        return pasm.pChar(pe.text)

    def PRange(self, pe, step):
        return pasm.pRange(pe.chars, pe.ranges)

    def PAnd(self, pe, step):
        e = self.inline(pe.e)
        if(self.Olex and isinstance(e, PChar)):
            return pasm.pAndChar(e.text)
        if(self.Olex and isinstance(e, PRange)):
            return pasm.pAndRange(e.chars, e.ranges)
        return pasm.pAnd(self.emit(e, step))

    def PNot(self, pe, step):
        e = self.inline(pe.e)
        if(self.Olex and isinstance(e, PChar)):
            return pasm.pNotChar(e.text)
        if(self.Olex and isinstance(e, PRange)):
            return pasm.pNotRange(e.chars, e.ranges)
        return pasm.pNot(self.emit(e, step))

    def PMany(self, pe, step):
        e = self.inline(pe.e)
        if(self.Olex and isinstance(e, PChar)):
            return pasm.pManyChar(e.text)
        if(self.Olex and isinstance(e, PRange)):
            return pasm.pManyRange(e.chars, e.ranges)
        return pasm.pMany(self.emit(e, step))

    def PMany1(self, pe, step):
        e = self.inline(pe.e)
        if(self.Olex and isinstance(e, PChar)):
            return pasm.pMany1Char(e.text)
        if(self.Olex and isinstance(e, PRange)):
            return pasm.pMany1Range(e.chars, e.ranges)
        return pasm.pMany1(self.emit(e, step))

    def POption(self, pe, step):
        e = self.inline(pe.e)
        if(self.Olex and isinstance(e, PChar)):
            return pasm.pOptionChar(e.text)
        if(self.Olex and isinstance(e, PRange)):
            return pasm.pOptionRange(e.chars, e.ranges)
        return pasm.pOption(self.emit(e, step))

    def PSeq(self, pe, step):
        pfs = []
        for e in pe:
            pfs.append(self.emit(e, step))
            step += e.minLen()
        pfs = tuple(pfs)
        if len(pfs) == 2:
            return pasm.pSeq2(pfs[0], pfs[1])
        if len(pe) == 3:
            return pasm.pSeq3(pfs[0], pfs[1], pfs[2])
        if len(pe) == 4:
            return pasm.pSeq4(pfs[0], pfs[1], pfs[2], pfs[3])
        return pasm.pSeq(*pfs)

    # Ore
    def POre(self, pe: POre, step):
        if pe.isDict():
            return pasm.pDict(pe.listDict())
        pfs = tuple(map(lambda e: self.emit(e, step), pe))
        if len(pfs) == 2:
            return pasm.pOre2(pfs[0], pfs[1])
        if len(pe) == 3:
            return pasm.pOre3(pfs[0], pfs[1], pfs[2])
        if len(pe) == 4:
            return pasm.pOre4(pfs[0], pfs[1], pfs[2], pfs[3])
        return pasm.pOre(*pfs)

    def PRef(self, pe, step):
        return pasm.pRef(self.generated, pe.uname())

    # Tree Construction

    def PNode(self, pe, step):
        _, fixed, es = self.fixedEach(0, [pe])
        # print(_, fixed, es)
        if fixed is None or not self.Ooox:
            fs = self.emit(pe.e, step)
            return pasm.pNode(fs, pe.tag, pe.shift)
        else:
            # print('//OOD', self.join(fixed, *es))
            return self.emit(self.join(fixed, *es), step)

    def PEdge(self, pe, step):
        return pasm.pEdge(pe.edge, self.emit(pe.e, step))

    def PFold(self, pe, step):
        _, fixed, es = self.fixedEach(0, [pe])
        # print(_, fixed, es)
        # fixed = None
        if fixed is None or not self.Ooox:
            fs = self.emit(pe.e, step)
            return pasm.pFold(pe.edge, fs, pe.tag, pe.shift)
        else:
            # print('//OOD', self.join(fixed, *es))
            return self.emit(self.join(fixed, *es), step)

    def PAbs(self, pe, step):
        return pasm.pAbs(self.emit(pe.e, step))

    def Skip(self, pe, step):  # @skip()
        return pasm.pSkip()

    def Symbol(self, pe, step):  # @symbol(A)
        params = pe.params
        sid = self.getsid(str(params[0]))
        return pasm.pSymbol(self.emit(pe.e, step), sid)

    def Scope(self, pe, step):
        return pasm.pScope(self.emit(pe.e, step))

    def Exists(self, pe, step):  # @Match(A)
        params = pe.params
        sid = self.getsid(str(params[0]))
        return pasm.pExists(sid)

    def Match(self, pe, step):  # @Match(A)
        params = pe.params
        sid = self.getsid(str(params[0]))
        return pasm.pMatch(sid)

    def Def(self, pe, step):  # @def(A, '名詞')
        params = pe.params
        name = str(params[1]) if len(params) == 2 else str(params[0])
        pf = self.emit(pe.e, step)
        #print('@def', pf, name)
        return pasm.pDef(name, pf)

    def In(self, pe, step):  # @in(A)
        params = pe.params
        name = str(params[0])
        #print('@in', name)
        return pasm.pIn(name)


generator = Generator()


def generate(peg, **options):
    return generator.generate(peg, **options)


# ParseTree


def logger(type, pos, msg):
    print(pos.showing(msg))


class TPEGLoader(object):
    def __init__(self, peg):
        self.names = {}
        self.peg = peg

    def load(self, t):
        for stmt in t:
            if stmt == 'Rule':
                name = str(stmt.name)
                if name in self.names:
                    # pos4 = stmt['name'].getpos4()
                    logger('error', stmt.name, f'redefined name {name}')
                    continue
                self.names[name] = stmt.e
            elif stmt == 'Example':
                doc = stmt.doc
                for n in stmt.names:
                    self.example(str(n), doc)
            # elif stmt == 'Import':
            #     urn = str(stmt.name)
            #     apeg = grammar(urn, **options)
            #     for n in stmt.names:
            #         lname = str(n)  # ns.Expression
            #         name = lname
            #         if lname.find('.') != -1:
            #             name = lname.split('.')[-1]
            #         pos4 = n.getpos4()
            #         if not name in apeg:
            #             logger('perror', pos4, f'undefined name {name}')
            #             continue
            #         g.add(lname, Action(apeg.newRef(name),
            #                             'import', (name, urn), pos4))
        for name in self.names:
            t = self.names[name]
            self.peg[name] = self.conv(t, 0)

    def example(self, name, doc):
        self.peg['@@example'].append((name, doc))

    def conv(self, t, step):
        tag = t.gettag()
        if hasattr(self, tag):
            f = getattr(self, tag)
            return f(t, step)
        return t

    @classmethod
    def unquote(cls, s):
        if s.startswith('\\'):
            if s.startswith('\\n'):
                return '\n', s[2:]
            if s.startswith('\\t'):
                return '\t', s[2:]
            if s.startswith('\\r'):
                return '\r', s[2:]
            if s.startswith('\\v'):
                return '\v', s[2:]
            if s.startswith('\\f'):
                return '\f', s[2:]
            if s.startswith('\\b'):
                return '\b', s[2:]
            if (s.startswith('\\x') or s.startswith('\\X')) and len(s) >= 4:
                c = int(s[2:4], 16)
                return chr(c), s[4:]
            if (s.startswith('\\u') or s.startswith('\\U')) and len(s) >= 6:
                c = chr(int(s[2:6], 16))
                if len(c) != 1:
                    c = ''  # remove unsupported unicode
                return c, s[6:]
            else:
                return s[1], s[2:]
        else:
            return s[0], s[1:]

    def Empty(self, t, step):
        return EMPTY

    def Any(self, t, step):
        return ANY

    def Char(self, t, step):
        s = str(t)
        sb = []
        while len(s) > 0:
            c, s = TPEGLoader.unquote(s)
            sb.append(c)
        return PChar(''.join(sb))

    def Class(self, t, step):
        s = str(t)
        chars = []
        ranges = []
        while len(s) > 0:
            c, s = TPEGLoader.unquote(s)
            if s.startswith('-') and len(s) > 1:
                c2, s = TPEGLoader.unquote(s[1:])
                ranges.append(c + c2)
            else:
                chars.append(c)
        if len(chars) == 0 and len(ranges) == 0:
            return EMPTY
        if len(chars) == 1 and len(ranges) == 0:
            return PChar(chars[0])
        return PRange(''.join(chars), ''.join(ranges))

    def Ref(self, t, step):
        name = str(t)
        if name in self.peg:
            return PAction(self.peg.newRef(name), 'NT', (name,), t.getpos4())
        if name[0].isupper() or name[0].islower() or name.startswith('_'):
            logger('warning', t, f'undefined nonterminal {name}')
            self.peg.add(name, EMPTY)
            return self.peg.newRef(name)
        return PChar(name[1:-1]) if name.startswith('"') else PChar(name)

    def Name(self, t, step):
        name = str(t)
        if name in self.names:
            return self.peg.newRef(name)
        if name[0].isupper() or name[0].islower() or name.startswith('_'):
            logger('warning', t, f'undefined nonterminal {name}')
            self.peg[name] = EMPTY
            return self.peg.newRef(name)
        return PChar(name[1:-1]) if name.startswith('"') else PChar(name)

    def Many(self, t, step):
        return PMany(self.conv(t.e, step))

    def Many1(self, t, step):
        return PMany1(self.conv(t.e, step))

    def Option(self, t, step):
        return POption(self.conv(t.e, step))

    def And(self, t, step):
        return PAnd(self.conv(t.e, step))

    def Not(self, t, step):
        return PNot(self.conv(t.e, step))

    def Seq(self, t, step):
        es = []
        for p in t:
            e = self.conv(p, step)
            # step += e.minLen()
            es.append(e)
        return PSeq.new(*es)

    def Ore(self, t, step):
        return POre.new(*tuple(map(lambda p: self.conv(p, step), t)))

    def Alt(self, t, step):
        return POre.new(*tuple(map(lambda p: self.conv(p, step), t)))

    def Node(self, t, step):
        tag = str(t.tag) if hasattr(t, 'tag') else ''
        e = self.conv(t.e, step)
        return PNode(e, tag, 0)

    def Edge(self, t, step):
        edge = str(t.edge) if hasattr(t, 'edge') else ''
        e = self.conv(t.e, step)
        return PEdge(edge, e)

    def Fold(self, t, step):
        edge = str(t.edge) if hasattr(t, 'edge') else ''
        tag = str(t.tag) if hasattr(t, 'tag') else ''
        e = self.conv(t.e, step)
        return PFold(edge, e, tag, 0)

    FIRST = {'lazy', 'scope', 'symbol', 'def',
             'match', 'equals', 'contains', 'cat'}

    def Func(self, t, step):
        funcname = str(t[0])
        ps = [self.conv(p, step) for p in t[1:]]
        if funcname.startswith('choice'):
            n = funcname[6:]
            if n.isdigit():
                return TPEGLoader.choiceN(t.urn_, int(n), ps)
            return TPEGLoader.choice(t.urn_, ps)
        if funcname in TPEGLoader.FIRST:
            return PAction(ps[0], funcname, tuple(ps), t)
        return PAction(EMPTY, funcname, tuple(ps), t)

    @classmethod
    def fileName(cls, e):
        s = str(e)
        return s[1:-1]  # if s.startswith('"') else s

    @classmethod
    def choice(cls, urn, es):
        ds = set()
        for e in es:
            file = TPEGLoader.fileName(e)
            file = Path(urn).parent / file
            with file.open(encoding='utf-8_sig') as f:
                ss = [x.strip('\r\n') for x in f.readlines()]
                ds |= {x for x in ss if len(x) > 0 and not x.startswith('#')}
        choice = [PChar(x) for x in sorted(ds, key=lambda x: len(x))[::-1]]
        return POre(*choice)

    @classmethod
    def choiceN(cls, urn, n, es):
        ds = set()
        for e in es:
            file = TPEGLoader.fileName(e)
            file = Path(urn).parent / file
            with file.open(encoding='utf-8_sig') as f:
                ss = [x.strip('\r\n') for x in f.readlines()]
                if n == 0:
                    ds |= {x for x in ss if len(
                        x) > 9 and not x.startswith('#')}
                else:
                    ds |= {x for x in ss if len(
                        x) == n and not x.startswith('#')}
        choice = [PChar(x) for x in ds]
        return POre(*choice)


def grammar_factory():

    def load_grammar(g, file, **options):
        # logger = options.get('logger', logger)
        # pegparser = pasm.generate(options.get('peg', TPEGGrammar))
        pegparser = pasm.generate(TPEGGrammar['Start'])
        if isinstance(file, Path):
            f = file.open(encoding=options.get('encoding', 'utf-8_sig'))
            data = f.read()
            f.close()
            t = pegparser(data, file)
            basepath = str(file)
        else:
            if 'basepath' in options:
                basepath = options['basepath']
            else:
                basepath = inspect.currentframe().f_back.f_code.co_filename
            t = pegparser(file, basepath)
            basepath = (str(Path(basepath).resolve().parent))
        options['basepath'] = basepath
        if t == 'err':
            logger('error', t, 'Syntax Error')
            return
        pconv = TPEGLoader(g)
        pconv.load(t)

    def findpath(paths, file):
        if file.find('=') > 0:
            return file
        for p in paths:
            path = Path(p) / file
            if path.exists():
                return path.resolve()
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file)

    GrammarDB = {}

    def grammar(urn, **options):
        paths = []
        basepath = options.get('basepath', '')
        if basepath == '':
            paths.append('')
        else:
            paths.append(str(Path(basepath).resolve().parent))
        framepath = inspect.currentframe().f_back.f_code.co_filename
        paths.append(str(Path(framepath).resolve().parent))
        paths.append(str(Path(__file__).resolve().parent / 'grammar'))
        paths += os.environ.get('GRAMMAR', '').split(':')
        path = findpath(paths, urn)
        key = str(path)
        if key in GrammarDB:
            return GrammarDB[key]
        peg = Grammar()
        load_grammar(peg, path, **options)
        GrammarDB[key] = peg
        return peg

    return grammar


grammar = grammar_factory()


if __name__ == '__main__':
    peg = grammar('es.tpeg')



import unittest
import tests
from pegpy.gparser.cython_gpeg import cgpeg
from pegpy.tpeg import grammar
from pathlib import Path

class TestCythonGPEG(unittest.TestCase):

  def test_exps(self):
    test_grammar = grammar('gpeg_grammar_test.gpeg')
    self.exTest(test_grammar, cgpeg)

if __name__ == '__main__':
  unittest.main()



#!/usr/local/bin/python
from pathlib import Path
import functools
import subprocess
# import readline
import time
import sys
import os
import importlib
import pegtree

istty = True


def bold(s):
    return '\033[1m' + str(s) + '\033[0m' if istty else str(s)


COLOR = {
    "Black": '0;30', "DarkGray": '1;30',
    "Red": '0;31', "LightRed": '1;31',
    "Green": '0;32', "LightGreen": '1;32',
    "Orange": '0;33', "Yellow": '1;33',
    "Blue": '0;34', "LightBlue": '1;34',
    "Purple": '0;35', "LightPurple": '1;35',
    "Cyan": '0;36', "LightCyan": '1;36',
    "LightGray": '0;37', "White": '1;37',
}


def color(c, s):
    return '\033[{}m{}\033[0m'.format(COLOR[c], str(s)) + '' if istty else str(s)


def showing(pos, msg):
    if pos is None:
        print(msg)
    else:
        print(pos.showing(msg))


def log(type, pos, *msg):
    msg = ' '.join(map(str, msg))
    if type.startswith('err'):
        showing(pos, color('Red', '[error] ') + str(msg))
    elif type.startswith('warn'):
        showing(pos, color('Orange', '[warning] ') + str(msg))
    elif type.startswith('info') or type.startswith('notice'):
        showing(pos, color('Cyan', '[info] ') + str(msg))
    else:
        showing(pos, str(msg))


def version():
    print(bold('PEGTree Parsing for Python3'))


def read_inputs(a):
    path = Path(a)
    if path.exists():
        f = path.open()
        data = f.read()
        f.close()
        return data
    else:
        return a


def readlines(prompt):
    s = input(prompt)
    if s != '':
        return s
    else:
        l = []
        while True:
            prev = s
            s = input()
            l.append(s)
            if prev == '' and s == '':
                break
        return '\n'.join(l)


def parse_options(argv):
    options = {
        'grammar': ['-g', '--grammar'],
        'start': ['-s', '--start'],
        'parser': ['-p', '--parser'],
        'output': ['-o', '--output'],
        'verbose': ['--verbose'],
    }

    def parse_each(a, d):
        if a[0].startswith('-'):
            if len(a) > 1:
                for key, list in options.items():
                    for l in list:
                        if a[0] == l:
                            d[key] = a[1]
                            return a[2:]
            d['inputs'].append(a)
            raise CommandUsageError
        else:
            d['inputs'].append(a[0])
            return a[1:]

    d = {'inputs': []}
    while len(argv) > 0:
        argv = parse_each(argv, d)
    d['logger'] = log
    return d


class CommandUsageError(Exception):
    pass


def usage():
    print("Usage: pegtree <command> options inputs")
    print("  -g | --grammar <file>      specify a grammar file")
    print("  -s | --start <NAME>        specify a starting rule")
    print("  -o | --output <file>       specify an output file")
    print("  -D                         specify an optional value")
    print()

    print("Example:")
    print("  pegtree parse -g math.tpeg <inputs>")
    print("  pegtree example -g math.tpeg <inputs>")
    print("  pegtree pasm -g math.tpeg")
    print()

    print("The most commonly used pegtree commands are:")
    print(" parse      run an interactive parser")
    print(" pasm       generate a parser combinator function")
    print(" example    test all examples")
    print(" update     update pegtree (via pip)")


def load_grammar(options, default=None):
    file = options.get('grammar', default)
    if file is None:
        raise CommandUsageError()
    if file == 'stdin.tpeg':
        data = sys.stdin.read()
        options['basepath'] = file
        return pegtree.grammar(data, **options)
    return pegtree.grammar(file, **options)


def generator(options):
    if 'parser' in options:
        m = importlib.import_module(options['parser'])
        return m.generate
    return pegtree.generate


# parse command


def parse(options, conv=None):
    peg = load_grammar(options)
    parser = generator(options)(peg, **options)
    inputs = options['inputs']
    if len(inputs) == 0:  # Interactive Mode
        try:
            while True:
                s = readlines(bold('>>> '))
                parser(s).dump(tag=lambda x: color('Blue', x))
        except (EOFError, KeyboardInterrupt):
            pass
    elif len(inputs) == 1:
        parser(read_inputs(inputs[0])).dump(tag=lambda x: color('Blue', x))
    else:
        for file in options['inputs']:
            st = time.time()
            t = parser(read_inputs(file))
            et = time.time()
            print(file, (et - st) * 1000.0, "[ms]:", t.tag)


def dump(t, indent='  ', edge=''):
    tag = color('Blue', '#' + t.tag)
    if t.child is None:
        s = t.inputs[t.spos: t.epos]
        print(indent + edge + bold("[")+tag, color('Red', repr(s)) + bold("]"))
        return
    print(indent + edge + bold("[") + tag)
    indent2 = '  ' + indent
    for tag, child in t.subs():
        if tag != '':
            tag = tag+'='
        dump(child, indent2, tag)
    print(indent + bold("]"))


def example(options):
    peg = load_grammar(options)
    if '@@example' not in peg:
        return
    parsers = {}
    for testcase in peg['@@example']:
        name, doc = testcase
        if not name in peg:
            continue
        if not name in parsers:
            parsers[name] = generator(options)(peg, start=name)
        res = parsers[name](doc.inputs_, doc.urn_, doc.spos_, doc.epos_)
        # print()
        ok = doc.inputs_[doc.spos_:res.epos_]
        fail = doc.inputs_[res.epos_:doc.epos_]
        print(bold(f'parsing {name}'), color(
            'Green', f'{ok}')+color('Red', f'{fail}'), bold('=> '), end='')
        res.dump()


def dumpError(lines, line, s):
    errs = 0
    for t in s:
        cur = str(t)
        if t.isSyntaxError():
            errs = 1
            s = max(0, t.spos_ - 10)
            prev = t.inputs_[s:t.spos_]
            # print(line)
            print(lines, color('Green', f'{prev}') + color('Red', f'{cur}'))
    return errs


def test(options):
    peg = load_grammar(options)
    parser = generator(options)(peg, **options)
    inputs = options['inputs']
    st = time.time()
    lines = 0
    fail = 0
    for file in options['inputs']:
        with open(file) as f:
            for line in f:
                lines += 1
                t = parser(line)
                fail += dumpError(lines, line, t)
    et = time.time()
    if lines > 0:
        print(f'{fail}/{lines} {fail/lines} {(et - st) * 1000.0} ms')


def peg(options):
    peg = load_grammar(options)
    print(peg)


def pasm0(options):
    from pegtree.parsec import parsec
    peg = load_grammar(options)
    options['optimized'] = 0
    parsec(peg, **options)


def pasm(options):
    from pegtree.parsec import parsec
    peg = load_grammar(options)
    parsec(peg, **options)


def update(options):
    try:
        # pip3 install -U git+https://github.com/KuramitsuLab/pegpy.git
        subprocess.check_call(
            ['pip3', 'install', '-U', 'pegtree'])
    except:
        pass


def update_beta(options):
    try:
        # pip3 install -U git+https://github.com/KuramitsuLab/pegpy.git
        subprocess.check_call(
            ['pip3', 'install', '-U', 'git+https://github.com/KuramitsuLab/pegtree.git'])
    except:
        pass


def main(argv=sys.argv):
    names = globals()
    if len(argv) > 1:
        cmd = argv[1]
        options = parse_options(argv[2:])
        cs = cmd.split('.')
        if len(cs) == 2:
            cmd = cs[0]
            options['ext'] = cs[1]
        if cmd in names:
            names[cmd](options)
            return
    usage()


if __name__ == "__main__":
    main(sys.argv)



from pegtree.pasm import *


def TPEG(peg):
    pRule(peg, 'Start', pSeq3(pRef(peg, '__'), pRef(
        peg, 'Source'), pRef(peg, 'EOF')))
    pRule(peg, '__', pMany(pOre2(pRange(' \t\r\n', ''), pRef(peg, 'COMMENT'))))
    pRule(peg, '_', pMany(pOre2(pRange(' \t', ''), pRef(peg, 'COMMENT'))))
    pRule(peg, 'COMMENT', pOre2(pSeq3(pChar('/*'), pMany(pSeq2(pNot(pChar('*/')), pAny())),
                                      pChar('*/')), pSeq2(pChar('//'), pMany(pSeq2(pNot(pRef(peg, 'EOL')), pAny())))))
    pRule(peg, 'EOL', pOre(pChar('\n'), pChar('\r\n'), pRef(peg, 'EOF')))
    pRule(peg, 'EOF', pNot(pAny()))
    pRule(peg, 'S', pRange(' \t', ''))
    pRule(peg, 'Source', pNode(
        pMany(pEdge('', pRef(peg, 'Statement'))), 'Source', 0))
    pRule(peg, 'EOS', pOre2(pSeq2(pRef(peg, '_'), pMany1(pSeq2(pChar(';'), pRef(
        peg, '_')))), pMany1(pSeq2(pRef(peg, '_'), pRef(peg, 'EOL')))))
    pRule(peg, 'Statement', pOre(pRef(peg, 'Import'), pRef(
        peg, 'Example'), pRef(peg, 'Rule')))
    pRule(peg, 'Import', pSeq2(pNode(pSeq(pChar('from'), pRef(peg, 'S'), pRef(peg, '_'), pEdge('name', pOre2(pRef(peg, 'Identifier'), pRef(peg, 'Char'))), pOption(
        pSeq(pRef(peg, '_'), pChar('import'), pRef(peg, 'S'), pRef(peg, '_'), pEdge('names', pRef(peg, 'Names'))))), 'Import', 0), pRef(peg, 'EOS')))
    pRule(peg, 'Example', pSeq2(pNode(pSeq(pChar('example'), pRef(peg, 'S'), pRef(peg, '_'), pEdge(
        'names', pRef(peg, 'Names')), pEdge('doc', pRef(peg, 'Doc'))), 'Example', 0), pRef(peg, 'EOS')))
    pRule(peg, 'Names', pNode(pSeq3(pEdge('', pRef(peg, 'Identifier')), pRef(peg, '_'), pMany(pSeq(
        pChar(','), pRef(peg, '_'), pEdge('', pRef(peg, 'Identifier')), pRef(peg, '_')))), '', 0))
    pRule(peg, 'Doc', pOre(pRef(peg, 'Doc1'),
                           pRef(peg, 'Doc2'), pRef(peg, 'Doc0')))
    pRule(peg, 'Doc0', pNode(pMany(pSeq2(pNot(pRef(peg, 'EOL')), pAny())), 'Doc', 0))
    pRule(peg, 'Doc1', pSeq(pRef(peg, 'DELIM1'), pMany(pRef(peg, 'S')), pRef(peg, 'EOL'), pNode(pMany(
        pSeq2(pNot(pSeq2(pRef(peg, 'DELIM1'), pRef(peg, 'EOL'))), pAny())), 'Doc', 0), pRef(peg, 'DELIM1')))
    pRule(peg, 'DELIM1', pChar("'''"))
    pRule(peg, 'Doc2', pSeq(pRef(peg, 'DELIM2'), pMany(pRef(peg, 'S')), pRef(peg, 'EOL'), pNode(pMany(
        pSeq2(pNot(pSeq2(pRef(peg, 'DELIM2'), pRef(peg, 'EOL'))), pAny())), 'Doc', 0), pRef(peg, 'DELIM2')))
    pRule(peg, 'DELIM2', pChar('```'))
    pRule(peg, 'Rule', pSeq2(pNode(pSeq(pEdge('name', pOre2(pRef(peg, 'Identifier'), pRef(peg, 'QName'))), pRef(peg, '__'), pOre2(pChar('='), pChar(
        '<-')), pRef(peg, '__'), pOption(pSeq2(pRange('/|', ''), pRef(peg, '__'))), pEdge('e', pRef(peg, 'Expression'))), 'Rule', 0), pRef(peg, 'EOS')))
    pRule(peg, 'Identifier', pNode(pRef(peg, 'NAME'), 'Name', 0))
    pRule(peg, 'NAME', pSeq2(pRange('_', 'AZaz'),
                             pMany(pRange('_.', 'AZaz09'))))
    pRule(peg, 'Expression', pSeq2(pRef(peg, 'Choice'), pOption(pFold('', pMany1(pSeq(pRef(peg, '__'), pChar(
        '|'), pNot(pChar('|')), pRef(peg, '_'), pEdge('', pRef(peg, 'Choice')))), 'Alt', 0))))
    pRule(peg, 'Choice', pSeq2(pRef(peg, 'Sequence'), pOption(pFold('', pMany1(pSeq(pRef(peg, '__'), pOre2(
        pChar('/'), pChar('||')), pRef(peg, '_'), pEdge('', pRef(peg, 'Sequence')))), 'Ore', 0))))
    pRule(peg, 'Sequence', pSeq2(pRef(peg, 'Predicate'), pOption(pFold('', pMany1(
        pSeq2(pRef(peg, 'SS'), pEdge('', pRef(peg, 'Predicate')))), 'Seq', 0))))
    pRule(peg, 'SS', pOre2(pSeq3(pRef(peg, 'S'), pRef(peg, '_'), pNot(pRef(peg, 'EOL'))), pSeq3(
        pMany1(pSeq2(pRef(peg, '_'), pRef(peg, 'EOL'))), pRef(peg, 'S'), pRef(peg, '_'))))
    pRule(peg, 'Predicate', pOre(pRef(peg, 'Not'), pRef(
        peg, 'And'), pRef(peg, 'Suffix')))
    pRule(peg, 'Not', pSeq2(pChar('!'), pNode(
        pEdge('e', pRef(peg, 'Predicate')), 'Not', 0)))
    pRule(peg, 'And', pSeq2(pChar('&'), pNode(
        pEdge('e', pRef(peg, 'Predicate')), 'And', 0)))
    pRule(peg, 'Suffix', pSeq2(pRef(peg, 'Term'), pOption(pOre(pFold('e', pChar(
        '*'), 'Many', 0), pFold('e', pChar('+'), 'Many1', 0), pFold('e', pChar('?'), 'Option', 0)))))
    pRule(peg, 'Term', pOre(pRef(peg, 'Group'), pRef(peg, 'Char'), pRef(peg, 'Class'), pRef(peg, 'Any'), pRef(
        peg, 'Node'), pRef(peg, 'Fold'), pRef(peg, 'EdgeFold'), pRef(peg, 'Edge'), pRef(peg, 'Func'), pRef(peg, 'Ref')))
    pRule(peg, 'Empty', pNode(pEmpty(), 'Empty', 0))
    pRule(peg, 'Group', pSeq(pChar('('), pRef(peg, '__'), pOre2(
        pRef(peg, 'Expression'), pRef(peg, 'Empty')), pRef(peg, '__'), pChar(')')))
    pRule(peg, 'Any', pNode(pChar('.'), 'Any', 0))
    pRule(peg, 'Char', pSeq3(pChar("'"), pNode(pMany(pOre2(pSeq2(
        pChar('\\'), pAny()), pSeq2(pNot(pChar("'")), pAny()))), 'Char', 0), pChar("'")))
    pRule(peg, 'Class', pSeq3(pChar('['), pNode(pMany(pOre2(pSeq2(
        pChar('\\'), pAny()), pSeq2(pNot(pChar(']')), pAny()))), 'Class', 0), pChar(']')))
    pRule(peg, 'Node', pNode(pSeq(pChar('{'), pRef(peg, '__'), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(peg, '__'))), pEdge('e', pOre2(pSeq2(pRef(
        peg, 'Expression'), pRef(peg, '__')), pRef(peg, 'Empty'))), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(peg, '__'))), pRef(peg, '__'), pChar('}')), 'Node', 0))
    pRule(peg, 'Tag', pSeq2(pChar('#'), pNode(
        pMany1(pSeq2(pNot(pRange(' \t\r\n}', '')), pAny())), 'Tag', 0)))
    pRule(peg, 'Fold', pNode(pSeq(pChar('^'), pRef(peg, '_'), pChar('{'), pRef(peg, '__'), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(peg, '__'))), pEdge('e', pOre2(pSeq2(
        pRef(peg, 'Expression'), pRef(peg, '__')), pRef(peg, 'Empty'))), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(peg, '__'))), pRef(peg, '__'), pChar('}')), 'Fold', 0))
    pRule(peg, 'Edge', pNode(pSeq(pEdge('edge', pRef(peg, 'Identifier')), pChar(':'), pRef(
        peg, '_'), pNot(pChar('^')), pEdge('e', pRef(peg, 'Term'))), 'Edge', 0))
    pRule(peg, 'EdgeFold', pNode(pSeq(pEdge('edge', pRef(peg, 'Identifier')), pChar(':'), pRef(peg, '_'), pChar('^'), pRef(peg, '_'), pChar('{'), pRef(peg, '__'), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(
        peg, '__'))), pEdge('e', pOre2(pSeq2(pRef(peg, 'Expression'), pRef(peg, '__')), pRef(peg, 'Empty'))), pOption(pSeq2(pEdge('tag', pRef(peg, 'Tag')), pRef(peg, '__'))), pRef(peg, '__'), pChar('}')), 'Fold', 0))
    pRule(peg, 'Func', pNode(pSeq(pChar('@'), pEdge('', pRef(peg, 'Identifier')), pChar('('), pRef(peg, '__'), pOre2(pEdge('', pRef(peg, 'Expression')), pEdge(
        '', pRef(peg, 'Empty'))), pMany(pSeq(pRef(peg, '_'), pChar(','), pRef(peg, '__'), pEdge('', pRef(peg, 'Expression')))), pRef(peg, '__'), pChar(')')), 'Func', 0))
    pRule(peg, 'Ref', pOre2(pRef(peg, 'Identifier'), pRef(peg, 'QName')))
    pRule(peg, 'QName', pNode(pSeq3(pChar('"'), pMany(pOre2(pSeq2(
        pChar('\\'), pAny()), pSeq2(pNot(pChar('"')), pAny()))), pChar('"')), 'Name', 0))
    return peg


TPEGGrammar = TPEG({})
# print(TPEGGrammar)
